"""FIFE: Finite-Interval Forecasting Engine.

Copyright (c) 2018 - 2021, Institute for Defense Analyses (IDA)

Please cite using the suggested citation in the LICENSE file.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details
(a copy is included in the LICENSE file).
"""

__version__ = "1.5.2"
